/* diversity-dhcp.c - DiversityDhcp
 *
 * Copyright 2008 Openmoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include "diversity-dhcp.h"
#include <diversity-ap.h>

#include <sys/types.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

enum {
	STATE_CHANGED,
	LAST_SIGNAL
};

enum {
	PROP_0,
	PROP_COMMAND_LINE,
};

enum _DiversityDhcpClient {
	DIVERSITY_DHCP_CLIENT_DHCLIENT,
	DIVERSITY_DHCP_CLIENT_UDHCPC,
};

enum _DiversityDhcpMessage {
	DIVERSITY_DHCP_MESSAGE_DISCOVER,
	DIVERSITY_DHCP_MESSAGE_OFFER,
	DIVERSITY_DHCP_MESSAGE_REQUEST,
	DIVERSITY_DHCP_MESSAGE_ACK,
	DIVERSITY_DHCP_MESSAGE_NAK,
	DIVERSITY_DHCP_MESSAGE_DECLINE,
	DIVERSITY_DHCP_MESSAGE_RELEASE,
	DIVERSITY_DHCP_MESSAGE_INFORM,

	DIVERSITY_DHCP_MESSAGE_UNKNOWN,
};

#if 0
enum _DiversityDhcpAction {
	DIVERSITY_DHCP_ACTION_MEDIUM,
	DIVERSITY_DHCP_ACTION_ARPCHECK,
	DIVERSITY_DHCP_ACTION_ARPSEND,
	DIVERSITY_DHCP_ACTION_PREINIT,
	DIVERSITY_DHCP_ACTION_BOUND,
	DIVERSITY_DHCP_ACTION_RENEW,
	DIVERSITY_DHCP_ACTION_REBBIND,
	DIVERSITY_DHCP_ACTION_EXPIRE,
	DIVERSITY_DHCP_ACTION_FAIL,
	DIVERSITY_DHCP_ACTION_RELEASE,
	DIVERSITY_DHCP_ACTION_TIMEOUT,
};
#endif

#define DIVERSITY_DHCP_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_DHCP, DiversityDhcpPrivate))
typedef struct _DiversityDhcpPrivate DiversityDhcpPrivate;

struct _DiversityDhcpPrivate {
	gchar *cmdline;
	gchar **cmdline_argv;
	gint cmdline_client;
	DiversityDhcpState state;

	GPid dhc_pid;
	gint dhc_fd;
	guint dhc_child_source;

	GIOChannel *io_channel;
	guint io_source;
};

static void diversity_dhcp_io_stop(DiversityDhcp *dhcpc);
static void dhc_kill(DiversityDhcpPrivate *priv);

static guint dhcp_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(DiversityDhcp, diversity_dhcp, G_TYPE_OBJECT);

static void diversity_dhcp_finalize(GObject *obj)
{
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(obj);

	diversity_dhcp_io_stop(DIVERSITY_DHCP(obj));
	dhc_kill(priv);

	g_strfreev(priv->cmdline_argv);
	g_free(priv->cmdline);

	((GObjectClass *) diversity_dhcp_parent_class)->finalize(obj);
}

static gboolean diversity_dhcp_set_cmdline(DiversityDhcp *dhcpc, const gchar *cmdline)
{
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);
	gboolean ret = FALSE;

	g_free(priv->cmdline);
	priv->cmdline = g_strdup(cmdline);

	priv->cmdline_client = DIVERSITY_DHCP_CLIENT_DHCLIENT;

	g_strfreev(priv->cmdline_argv);
	if (!g_shell_parse_argv(priv->cmdline,
				NULL, &priv->cmdline_argv, NULL)) {
		g_warning("failed to parse %s", priv->cmdline);

		priv->cmdline_argv = NULL;

		return FALSE;
	}

	if (priv->cmdline_argv) {
		const gchar *argv0;

		argv0 = *priv->cmdline_argv;
		if (argv0 && g_file_test(argv0, G_FILE_TEST_IS_EXECUTABLE)) {
			if (strstr(argv0, "udhcpc"))
				priv->cmdline_client =
					DIVERSITY_DHCP_CLIENT_UDHCPC;

			ret = TRUE;
		} else {
			g_strfreev(priv->cmdline_argv);
			priv->cmdline_argv = NULL;
		}
	}

	return ret;
}

static void diversity_dhcp_set_property(GObject *obj, guint prop_id, const GValue *value, GParamSpec *pspec)
{
	DiversityDhcp *dhcpc = DIVERSITY_DHCP(obj);

	switch (prop_id) {
	case PROP_COMMAND_LINE:
		diversity_dhcp_set_cmdline(dhcpc,
				g_value_get_string(value));
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_dhcp_get_property(GObject *obj, guint prop_id, GValue *value, GParamSpec *pspec)
{
	DiversityDhcp *dhcpc = DIVERSITY_DHCP(obj);
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);

	switch (prop_id) {
	case PROP_COMMAND_LINE:
		g_value_set_string(value, priv->cmdline);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID(obj, prop_id, pspec);
		break;
	}
}

static void diversity_dhcp_class_init(DiversityDhcpClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;

	o_class->finalize = diversity_dhcp_finalize;
	o_class->set_property = diversity_dhcp_set_property;
	o_class->get_property = diversity_dhcp_get_property;

	dhcp_signals[STATE_CHANGED] =
		g_signal_new("state-changed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityDhcpClass, state_changed),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__INT,
			     G_TYPE_NONE, 1,
			     G_TYPE_INT);

	g_object_class_install_property(o_class,
					PROP_COMMAND_LINE,
					g_param_spec_string("command-line",
						 	    "Command line",
							    "Command line",
							    "",
							    G_PARAM_READWRITE |
							    G_PARAM_CONSTRUCT));

	g_type_class_add_private(klass, sizeof(DiversityDhcpPrivate));
}

static void diversity_dhcp_init(DiversityDhcp *dhcpc)
{
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);

	priv->cmdline = NULL;
	priv->cmdline_client = DIVERSITY_DHCP_CLIENT_DHCLIENT;
	priv->state = DIVERSITY_DHCP_STATE_SLEEP;

	priv->dhc_pid = -1;
	priv->dhc_fd = -1;
}

DiversityDhcp *diversity_dhcp_new(const gchar *cmdline)
{
	DiversityDhcp *dhcpc;

	dhcpc = g_object_new(DIVERSITY_TYPE_DHCP,
			    "command-line", cmdline,
			    NULL);

	return dhcpc;
}

DiversityDhcpState diversity_dhcp_get_state(DiversityDhcp *dhcpc)
{
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);

	return priv->state;
}

/* this function is called in child process */
static void dhc_child_setup(gpointer data)
{
	setpgid(0, 0);
}

static gboolean dhc_spawn(DiversityDhcpPrivate *priv, GChildWatchFunc watch_func, gpointer watch_data, GError **error)
{
	gint *pstdout, *pstderr;

	if (!priv->cmdline_argv) {
		g_set_error(error, 0, 0, "no command line");

		return FALSE;
	}

	if (priv->cmdline_client == DIVERSITY_DHCP_CLIENT_UDHCPC) {
		pstdout = &priv->dhc_fd;
		pstderr = NULL;
	} else {
		pstdout = NULL;
		pstderr = &priv->dhc_fd;
	}

	if (!g_spawn_async_with_pipes("/",
			   priv->cmdline_argv,
			   NULL,
			   G_SPAWN_DO_NOT_REAP_CHILD,
			   dhc_child_setup,
			   priv,
			   &priv->dhc_pid,
			   NULL,
			   pstdout,
			   pstderr,
			   error))
		return FALSE;

	priv->dhc_child_source = g_child_watch_add(priv->dhc_pid, watch_func, watch_data);

	return TRUE;
}

static void dhc_release(DiversityDhcpPrivate *priv)
{
	if (priv->dhc_child_source) {
		g_source_remove(priv->dhc_child_source);
		priv->dhc_child_source = 0;
	}

	if (priv->dhc_fd >= 0) {
		close(priv->dhc_fd);
		priv->dhc_fd = -1;
	}

	if (priv->dhc_pid >= 0) {
		g_spawn_close_pid(priv->dhc_pid);
		priv->dhc_pid = -1;
	}
}

static void dhc_kill(DiversityDhcpPrivate *priv)
{
	if (priv->dhc_pid < 0)
		return;

	kill(priv->dhc_pid, SIGTERM);
	dhc_release(priv);
}

static void diversity_dhcp_set_state(DiversityDhcp *dhcpc, DiversityDhcpState state)
{
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);

	if (priv->state != state) {
		priv->state = state;
		g_signal_emit(dhcpc, dhcp_signals[STATE_CHANGED], 0,
				state);
	}
}

static void diversity_dhcp_io_parse_udhcpc(DiversityDhcp *dhcpc, gchar *line)
{
	gchar *p;

	if (*line == 'S' && strncmp(line, "Sending ", 8) == 0) {
		line += 8;

		p = line;
		while (isalpha(*p))
			p++;
		if (*p)
			*p++ = '\0';

		if (strcmp(line, "discover") == 0)
			diversity_dhcp_set_state(dhcpc, DIVERSITY_DHCP_STATE_SELECTING);
		else if (strcmp(line, "select") == 0)
			diversity_dhcp_set_state(dhcpc, DIVERSITY_DHCP_STATE_REQUESTING);
	} else if (*line == 'L' && strncmp(line, "Lease ", 6) == 0) {
		line += 6;

		p = line;
		while (isalpha(*p))
			p++;
		if (*p)
			*p++ = '\0';

		if (strcmp(line, "failed") == 0)
			diversity_dhcp_set_state(dhcpc, DIVERSITY_DHCP_STATE_SLEEP);
		else if (strcmp(line, "of") == 0)
			diversity_dhcp_set_state(dhcpc, DIVERSITY_DHCP_STATE_BOUND);
	} else {
		if ((*line == 'a' && strncmp(line, "adding dns", 10) == 0) ||
				strstr(line, "route"))
			diversity_dhcp_set_state(dhcpc, DIVERSITY_DHCP_STATE_CONFIGURED);
	}
}

static gint get_message_type(const gchar *msg)
{
	gint msg_type;

	if (strcmp(msg, "DISCOVER") == 0)
		msg_type = DIVERSITY_DHCP_MESSAGE_DISCOVER;
	else if (strcmp(msg, "OFFER") == 0)
		msg_type = DIVERSITY_DHCP_MESSAGE_OFFER;
	else if (strcmp(msg, "REQUEST") == 0)
		msg_type = DIVERSITY_DHCP_MESSAGE_REQUEST;
	else if (strcmp(msg, "ACK") == 0)
		msg_type = DIVERSITY_DHCP_MESSAGE_ACK;
	else if (strcmp(msg, "NAK") == 0)
		msg_type = DIVERSITY_DHCP_MESSAGE_NAK;
	else if (strcmp(msg, "DECLINE") == 0)
		msg_type = DIVERSITY_DHCP_MESSAGE_DECLINE;
	else if (strcmp(msg, "RELEASE") == 0)
		msg_type = DIVERSITY_DHCP_MESSAGE_RELEASE;
	else if (strcmp(msg, "INFORM") == 0)
		msg_type = DIVERSITY_DHCP_MESSAGE_INFORM;
	else
		msg_type = DIVERSITY_DHCP_MESSAGE_UNKNOWN;

	return msg_type;
}

static void diversity_dhcp_io_parse_dhclient(DiversityDhcp *dhcpc, gchar *line)
{
	gchar *p;
	gint msg_type;

	if (*line != 'D' || strncmp(line, "DHCP", 4) != 0) {
		if (*line == 'b' && strncmp(line, "bound to ", 9) == 0)
			diversity_dhcp_set_state(dhcpc,
					DIVERSITY_DHCP_STATE_CONFIGURED);
		else if (*line == 'N' && strstr(line, "sleeping"))
			diversity_dhcp_set_state(dhcpc,
					DIVERSITY_DHCP_STATE_SLEEP);

		return;
	}

	line += 4;
	p = strchr(line, ' ');
	if (!p)
		return;
	*p++ = '\0';

	msg_type = get_message_type(line);

	/* simple state transition */
	switch (msg_type) {
	case DIVERSITY_DHCP_MESSAGE_DISCOVER:
	case DIVERSITY_DHCP_MESSAGE_OFFER:
		diversity_dhcp_set_state(dhcpc, DIVERSITY_DHCP_STATE_SELECTING);
		break;
	case DIVERSITY_DHCP_MESSAGE_REQUEST:
	case DIVERSITY_DHCP_MESSAGE_NAK:
	case DIVERSITY_DHCP_MESSAGE_DECLINE:
		diversity_dhcp_set_state(dhcpc, DIVERSITY_DHCP_STATE_REQUESTING);
		break;
	case DIVERSITY_DHCP_MESSAGE_ACK:
	case DIVERSITY_DHCP_MESSAGE_INFORM:
		diversity_dhcp_set_state(dhcpc, DIVERSITY_DHCP_STATE_BOUND);
		break;
	case DIVERSITY_DHCP_MESSAGE_RELEASE:
		diversity_dhcp_set_state(dhcpc, DIVERSITY_DHCP_STATE_INIT);
		break;
	case DIVERSITY_DHCP_MESSAGE_UNKNOWN:
	default:
		g_debug("received DHCP%s", line);
		break;
	}
}

static gboolean io_input(GIOChannel *channel, GIOCondition condition, gpointer data)
{
	DiversityDhcp *dhcpc = data;
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);
	gchar *line;
	gsize len;
	GIOStatus status;
	gchar *message;

	switch (condition) {
	case G_IO_IN:
	case G_IO_PRI:
		/* FIXME this is heavy */
		status = g_io_channel_read_line(priv->io_channel, &line, &len, NULL, NULL);
		if (status == G_IO_STATUS_AGAIN) {
			break;
		} else if (status != G_IO_STATUS_NORMAL) {
			message = "failed to read";

			goto fail;
		}

		switch (priv->cmdline_client)
		{
		case DIVERSITY_DHCP_CLIENT_UDHCPC:
			diversity_dhcp_io_parse_udhcpc(dhcpc, line);
			break;
		case DIVERSITY_DHCP_CLIENT_DHCLIENT:
		default:
			diversity_dhcp_io_parse_dhclient(dhcpc, line);
			break;
		}
		g_free(line);
		break;
	case G_IO_ERR:
	case G_IO_HUP:
		message = "connection broken";

		goto fail;
		break;
	default:
		message = "unknown IO condition";

		goto fail;
		break;
	}

	return TRUE;

fail:
	g_warning("io_input failed(%d): %s", condition, message);

	g_io_channel_unref(priv->io_channel);
	priv->io_channel = NULL;
	priv->io_source = 0;

	return FALSE;
}

static void diversity_dhcp_io_start(DiversityDhcp *dhcpc)
{
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);

	if (priv->io_channel)
		return;

	priv->io_channel = g_io_channel_unix_new(priv->dhc_fd);
	g_io_channel_set_encoding(priv->io_channel, NULL, NULL);

	priv->io_source = g_io_add_watch(priv->io_channel,
			G_IO_IN | G_IO_PRI | G_IO_ERR | G_IO_HUP,
			io_input, dhcpc);
}

static void diversity_dhcp_io_stop(DiversityDhcp *dhcpc)
{
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);

	if (priv->io_channel) {
		g_io_channel_unref(priv->io_channel);
		priv->io_channel = NULL;

		g_source_remove(priv->io_source);
		priv->io_source = 0;
	}
}

static void diversity_dhcp_child_watch(GPid pid, gint status, gpointer data)
{
	DiversityDhcp *dhcpc = data;
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);

	g_debug("child %d died: %d", pid, status);

	if (priv->dhc_pid == pid) {
		diversity_dhcp_io_stop(dhcpc);
		dhc_release(priv);

		diversity_dhcp_set_state(dhcpc,
				DIVERSITY_DHCP_STATE_SLEEP);
	}
}

gboolean diversity_dhcp_wakeup(DiversityDhcp *dhcpc, GError **error)
{
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);

	if (priv->state != DIVERSITY_DHCP_STATE_SLEEP)
		return TRUE;

	dhc_kill(priv);

	if (!dhc_spawn(priv, diversity_dhcp_child_watch, dhcpc, error))
		return FALSE;

	diversity_dhcp_io_start(dhcpc);

	diversity_dhcp_set_state(dhcpc,
			DIVERSITY_DHCP_STATE_INIT);

	return TRUE;
}

void diversity_dhcp_sleep(DiversityDhcp *dhcpc)
{
	DiversityDhcpPrivate *priv = DIVERSITY_DHCP_GET_PRIVATE(dhcpc);

	if (priv->state != DIVERSITY_DHCP_STATE_SLEEP) {
		diversity_dhcp_io_stop(dhcpc);
		dhc_kill(priv);

		diversity_dhcp_set_state(dhcpc,
				DIVERSITY_DHCP_STATE_SLEEP);
	}
}
