/* diversity-rae.c - DiversityRae
 *
 * Copyright 2008 Openmoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <glib.h>
#include <diversity-equipment.h>
#include <diversity-world.h>
#include <diversity-net.h>
#include <diversity-ap.h>
#include <string.h>

#include <dbus/dbus-glib.h>
#include "rae-marshal.h"

#include "diversity-rae.h"
#include "diversity-apdb.h"

enum {
	STATUS_CHANGED,
	QUERY_DONE,
	REPORT_DONE,
	LAST_SIGNAL
};

#define DIVERSITY_RAE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_RAE, DiversityRaePrivate))
typedef struct _DiversityRaePrivate DiversityRaePrivate;

enum {
	DIVERSITY_RAE_REQUEST_LOGIN,
	DIVERSITY_RAE_REQUEST_QUERY,
	DIVERSITY_RAE_REQUEST_REPORT,
	DIVERSITY_RAE_NUM_REQUESTS
};

struct _DiversityRaePrivate {
	DiversityWorld *world;

	DiversityApdb *apdb;
	gchar *server;

	DiversityNet *net;
	gboolean net_activated;

	guint32 serial_number;
	guint32 requests[DIVERSITY_RAE_NUM_REQUESTS];
};

static gboolean control_rae_get_status(DiversityRae *rae, gchar **OUT_status, GError **error);
static gboolean control_rae_request_login(DiversityRae *rae, const gchar *IN_username, const gchar *IN_password, guint *OUT_id, GError **error);
static gboolean control_rae_request_query(DiversityRae *rae, gdouble IN_lon, gdouble IN_lat, gdouble IN_radius, guint *OUT_id, GError **error);
static gboolean control_rae_request_report(DiversityRae *rae, const gchar *IN_ap, guint *OUT_id, GError **error);
static gboolean control_rae_cancel_request(DiversityRae *rae, guint IN_id, GError **error);
#include "diversity-rae-glue.h"
/* XXX control is core! */
#include "../diversity-control-backend/diversity-control.h"

static guint rae_signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(DiversityRae, diversity_rae, DIVERSITY_TYPE_EQUIPMENT);

static void diversity_rae_dispose(GObject *obj)
{
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(obj);

	if (priv->apdb) {
		g_object_unref(priv->apdb);
		priv->apdb = NULL;
	}

	if (priv->net) {
		g_object_unref(priv->net);
		priv->net = NULL;
	}

	if (priv->world) {
		g_object_unref(priv->world);
		priv->world = NULL;
	}

	((GObjectClass *) diversity_rae_parent_class)->dispose(obj);
}

static void diversity_rae_finalize(GObject *obj)
{
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(obj);

	g_free(priv->server);

	((GObjectClass *) diversity_rae_parent_class)->finalize(obj);
}

static void on_net_state_changed(DiversityNet *net, DiversityNetState state, gpointer user_data)
{
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(user_data);

	priv->net_activated = (state == DIVERSITY_NET_STATE_ACTIVATED);
}

static gboolean diversity_rae_equip(DiversityEquipment *eqp)
{
	DiversityRae *rae = DIVERSITY_RAE(eqp);
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);
	DiversityNetState state;

	priv->net = diversity_world_net_get(priv->world);
	if (!priv->net) {
		g_warning("no DiversityNet provider");

		return TRUE;
	}

	g_object_ref(priv->net);

	g_signal_connect(priv->net, "state-changed",
			G_CALLBACK(on_net_state_changed), rae);

	state = diversity_net_get_state(priv->net);
	on_net_state_changed(priv->net, state, rae);

	return TRUE;
}

static void diversity_rae_unequip(DiversityEquipment *eqp)
{
	DiversityRae *rae = DIVERSITY_RAE(eqp);
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);

	if (priv->net) {
		on_net_state_changed(priv->net,
				DIVERSITY_NET_STATE_UNKNOWN, rae);

		g_signal_handlers_disconnect_by_func(priv->net,
				on_net_state_changed, rae);

		g_object_unref(priv->net);

		priv->net = NULL;
	}
}

static gboolean diversity_rae_set_config(DiversityEquipment *eqp, const gchar *key, const GValue *val)
{
	return FALSE;
}

static gboolean diversity_rae_get_config(DiversityEquipment *eqp, const gchar *key, GValue *val)
{
	return FALSE;
}

static void diversity_rae_class_init(DiversityRaeClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;
	DiversityEquipmentClass *eqp_class = (DiversityEquipmentClass *) klass;

	o_class->dispose = diversity_rae_dispose;
	o_class->finalize = diversity_rae_finalize;

	eqp_class->equip = diversity_rae_equip;
	eqp_class->unequip = diversity_rae_unequip;

	eqp_class->set_config = diversity_rae_set_config;
	eqp_class->get_config = diversity_rae_get_config;

	g_type_class_add_private(klass, sizeof(DiversityRaePrivate));

	rae_signals[STATUS_CHANGED] =
		g_signal_new("status-changed",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityRaeClass, status_changed),
			     NULL, NULL,
			     g_cclosure_marshal_VOID__STRING,
			     G_TYPE_NONE, 1,
			     G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);

	rae_signals[QUERY_DONE] =
		g_signal_new("query-done",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityRaeClass, query_done),
			     NULL, NULL,
			     rae_marshal_VOID__UINT_STRING,
			     G_TYPE_NONE, 2,
			     G_TYPE_UINT,
			     G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);

	rae_signals[REPORT_DONE] =
		g_signal_new("report-done",
			     G_TYPE_FROM_CLASS(klass),
			     G_SIGNAL_RUN_LAST,
			     G_STRUCT_OFFSET(DiversityRaeClass, report_done),
			     NULL, NULL,
			     rae_marshal_VOID__UINT_STRING,
			     G_TYPE_NONE, 2,
			     G_TYPE_UINT,
			     G_TYPE_STRING | G_SIGNAL_TYPE_STATIC_SCOPE);

	dbus_g_object_type_install_info(G_TYPE_FROM_CLASS(klass),
			&dbus_glib_control_rae_object_info);
}

static void diversity_rae_init(DiversityRae *rae)
{
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);

	priv->server = g_strdup("https://diversity.openmoko.net/php/");
}

static void ap_position(DiversityAp *ap, gdouble lon, gdouble lat)
{
	gdouble len;

	len = DIVERSITY_UNITS(200);
	lon -= len / 2;
	lat -= len / 2;

	diversity_object_geometry_set(DIVERSITY_OBJECT(ap),
			lon, lat, len, len);
}

static void on_query_done(DiversityApdb *apdb, GSList *aps, gpointer data)
{
	DiversityRae *rae = DIVERSITY_RAE(data);
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);

	if (priv->requests[DIVERSITY_RAE_REQUEST_QUERY]) {
		guint32 id;

		id = priv->requests[DIVERSITY_RAE_REQUEST_QUERY];
		priv->requests[DIVERSITY_RAE_REQUEST_QUERY] = 0;

		g_signal_emit(rae, rae_signals[QUERY_DONE], 0, id,
				(aps) ? "succeeded" : "failed");
	}

	while (aps) {
		DiversityAp *ap;
		ApDesc *desc = aps->data;

		if (!desc->hw_addr)
			continue;

		ap = (DiversityAp *)
			diversity_world_lookup_object(priv->world,
				DIVERSITY_OBJECT_TYPE_AP, desc->hw_addr);
		if (!ap) {
			ap = diversity_ap_new();
			DIVERSITY_OBJECT(ap)->id = g_strdup(desc->hw_addr);
			diversity_object_set_persistent(DIVERSITY_OBJECT(ap), TRUE);

			diversity_ap_set_hw_address(ap, desc->hw_addr);
			diversity_ap_set_ssid(ap, desc->ssid);
			ap_position(ap, desc->lon, desc->lat);

			diversity_world_add_object(priv->world, DIVERSITY_OBJECT(ap));

			g_debug("new AP %s added", desc->ssid);
		}

		aps = aps->next;
	}
}

static void on_update_done(DiversityApdb *apdb, gboolean success, gpointer data)
{
	DiversityRae *rae = DIVERSITY_RAE(data);
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);

	if (priv->requests[DIVERSITY_RAE_REQUEST_REPORT]) {
		guint32 id;

		id = priv->requests[DIVERSITY_RAE_REQUEST_REPORT];
		priv->requests[DIVERSITY_RAE_REQUEST_REPORT] = 0;

		g_signal_emit(rae, rae_signals[REPORT_DONE], 0, id,
				(success) ? "succeeded" : "failed");
	}
}

static void on_ready(DiversityApdb *apdb, guint version, gpointer data)
{
	DiversityRae *rae = DIVERSITY_RAE(data);
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);
	const char *status;
	gboolean ready;

	ready = (version > 0);

	status = (ready) ? "connected" : "disconnected";
	g_debug("server %s", status);

	if (priv->requests[DIVERSITY_RAE_REQUEST_LOGIN]) {
		priv->requests[DIVERSITY_RAE_REQUEST_LOGIN] = 0;
		g_signal_emit(rae, rae_signals[STATUS_CHANGED], 0, status);
	}
}

DiversityRae *diversity_rae_new(DiversityWorld *world, GError **error)
{
	DiversityRae *rae;
	DiversityRaePrivate *priv;

	rae = g_object_new(DIVERSITY_TYPE_RAE,
			    "name", "rae",
			    NULL);

	priv = DIVERSITY_RAE_GET_PRIVATE(rae);

	priv->world = world;
	g_object_ref(priv->world);

	priv->apdb = diversity_apdb_new(NULL);
	g_signal_connect(priv->apdb, "update-done", G_CALLBACK(on_update_done), rae);
	g_signal_connect(priv->apdb, "query-done", G_CALLBACK(on_query_done), rae);
	g_signal_connect(priv->apdb, "ready", G_CALLBACK(on_ready), rae);

	return rae;
}

/* Implementation of the DBus interface */

static gboolean control_rae_get_status(DiversityRae *rae, gchar **OUT_status, GError **error)
{
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);
	const char *status;

	status = (diversity_apdb_is_ready(priv->apdb))
		? "connected" : "disconnected";

	*OUT_status = g_strdup(status);

	return TRUE;
}

static gboolean control_rae_request_login(DiversityRae *rae, const gchar *IN_username, const gchar *IN_password, guint *OUT_id, GError **error)
{
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);

	if (priv->requests[DIVERSITY_RAE_REQUEST_LOGIN]) {
		g_set_error(error, DIVERSITY_CONTROL_ERROR, 0,
				"queue full");

		return FALSE;
	}

	if (diversity_apdb_is_ready(priv->apdb)) {
		diversity_apdb_cancel_query(priv->apdb);
		diversity_apdb_cancel_update(priv->apdb);

		g_signal_emit(rae, rae_signals[STATUS_CHANGED], 0,
				"connected");

		return TRUE;
	}

	diversity_apdb_set_server(priv->apdb, priv->server);

	*OUT_id = ++priv->serial_number;
	priv->requests[DIVERSITY_RAE_REQUEST_LOGIN] = *OUT_id;

	return TRUE;
}

static gboolean control_rae_request_query(DiversityRae *rae, gdouble IN_lon, gdouble IN_lat, gdouble IN_radius, guint *OUT_id, GError **error)
{
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);

	if (priv->requests[DIVERSITY_RAE_REQUEST_QUERY]) {
		g_set_error(error, DIVERSITY_CONTROL_ERROR, 0,
				"queue full");

		return FALSE;
	}

	if (!diversity_apdb_is_ready(priv->apdb)) {
		g_set_error(error, DIVERSITY_CONTROL_ERROR, 0,
				"no connection");

		return FALSE;
	}

	if (!diversity_apdb_request_query(priv->apdb,
				IN_lon, IN_lat, IN_radius)) {
		g_set_error(error, DIVERSITY_CONTROL_ERROR, 0,
				"failed to send request");

		return FALSE;
	}

	*OUT_id = ++priv->serial_number;
	priv->requests[DIVERSITY_RAE_REQUEST_QUERY] = *OUT_id;

	return TRUE;
}

static gboolean describe_ap(ApDesc *desc, DiversityAp *ap)
{
	gdouble w, h;

	diversity_object_geometry_get(DIVERSITY_OBJECT(ap),
			&desc->lon, &desc->lat, &w, &h);
	desc->lon += w / 2;
	desc->lat += h / 2;

	desc->hw_addr = diversity_ap_get_hw_address(ap);
	desc->ssid = diversity_ap_get_ssid(ap);

	return TRUE;
}

static gboolean control_rae_request_report(DiversityRae *rae, const gchar *IN_ap, guint *OUT_id, GError **error)
{
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);
	DiversityControl *ctrl;
	ApDesc ap;
	DiversityObject *obj = NULL;

	if (priv->requests[DIVERSITY_RAE_REQUEST_REPORT]) {
		g_set_error(error, DIVERSITY_CONTROL_ERROR, 0,
				"queue full");

		return FALSE;
	}

	if (!diversity_apdb_is_ready(priv->apdb)) {
		g_set_error(error, DIVERSITY_CONTROL_ERROR, 0,
				"no connection");

		return FALSE;
	}

	ctrl = diversity_control_from_equipment(DIVERSITY_EQUIPMENT(rae));
	if (ctrl)
		obj = diversity_control_object_lookup(ctrl, IN_ap);
	if (!ctrl || !obj || !DIVERSITY_IS_AP(obj)) {
		g_set_error(error, DIVERSITY_CONTROL_ERROR, 0,
				"invalid argument");

		return FALSE;
	}

	describe_ap(&ap, DIVERSITY_AP(obj));

	if (!diversity_apdb_request_update(priv->apdb, &ap, 1)) {
		g_set_error(error, DIVERSITY_CONTROL_ERROR, 0,
				"failed to send request");

		return FALSE;
	}

	*OUT_id = ++priv->serial_number;
	priv->requests[DIVERSITY_RAE_REQUEST_REPORT] = *OUT_id;

	return TRUE;
}

static gboolean control_rae_cancel_request(DiversityRae *rae, guint IN_id, GError **error)
{
	DiversityRaePrivate *priv = DIVERSITY_RAE_GET_PRIVATE(rae);
	gint i;

	if (!IN_id)
		return TRUE;

	for (i = 0; i < DIVERSITY_RAE_NUM_REQUESTS; i++) {
		if (IN_id == priv->requests[i])
			break;
	}

	if (i >= DIVERSITY_RAE_NUM_REQUESTS)
		return TRUE;

	switch (i) {
	case DIVERSITY_RAE_REQUEST_LOGIN:
		diversity_apdb_set_server(priv->apdb, NULL);
		break;
	case DIVERSITY_RAE_REQUEST_QUERY:
		diversity_apdb_cancel_query(priv->apdb);
		break;
	case DIVERSITY_RAE_REQUEST_REPORT:
		diversity_apdb_cancel_update(priv->apdb);
		break;
	}

	return TRUE;
}
