/* diversity-backend - 
 *
 * Copyright 2007-2008 Openmoko, Inc.
 * Authored by Daniel Willmann <daniel@openmoko.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <stdlib.h>
#include <signal.h>
#include <syslog.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <diversity-world.h>
#include <diversity-module.h>

#include <string.h> /* for strcmp */

#include "builtin-backends.h"

static GMainLoop *mainloop;
GSList *loaded_modules;

static void on_signal(int sig)
{
	static gboolean quitting = FALSE;

	if (quitting)
		return;

	quitting = TRUE;

	g_warning("caught signal %d", sig);

	if (mainloop)
		g_main_loop_quit(mainloop);
	else
		exit(1);
}

static gboolean no_daemon;

static GOptionEntry entries[] = {
	{ "no-daemon", 0, 0, G_OPTION_ARG_NONE, &no_daemon, "Don't become a daemon", NULL },
	{ NULL }
};

static int init(int argc, char **argv, GError **error)
{
	struct sigaction act;
	GOptionContext *context;

	act.sa_handler = on_signal;
	sigemptyset(&act.sa_mask);
	act.sa_flags = 0;

	sigaddset(&act.sa_mask, SIGTERM);
	sigaction(SIGINT, &act, NULL);
	sigaction(SIGTERM, &act, NULL);

	context = g_option_context_new(NULL);
	g_option_context_add_main_entries(context, entries, NULL);

	if (!g_option_context_parse(context, &argc, &argv, error))
		return FALSE;

	g_option_context_free(context);

	return TRUE;
}

static gboolean backends_load_builtins(DiversityWorld *world, GError **error)
{
	DiversityModuleHandle * const * builtins;

	for (builtins = _diversity_builtin_backends; *builtins; builtins++) {
		DiversityModuleHandle *mh = *builtins;
	       
		if (!mh->name)
			continue;

		g_debug("initialize %s backend", mh->name);

		if (mh->init(world, error)) {
			DiversityModule *mod =
				diversity_module_new_from_handle(mh);

			loaded_modules = g_slist_prepend(loaded_modules, mod);
		} else {
			if (strcmp(mh->name, "Control") == 0)
				return FALSE;

			g_warning("failed to initialize %s backend: %s",
					mh->name, (*error)->message);

			g_clear_error(error);
		}
	}

	return TRUE;
}

static gboolean backends_load(DiversityWorld *world, GError **error)
{
	return backends_load_builtins(world, error);
}

static void backends_unload(DiversityWorld *world)
{
	while (loaded_modules) {
		DiversityModule *mod = loaded_modules->data;
		DiversityModuleHandle *mh = diversity_module_get_handle(mod);

		mh->fini();
		g_object_unref(mod);

		loaded_modules =
			g_slist_delete_link(loaded_modules, loaded_modules);
	}
}

static void equip_equipments(DiversityWorld *world, gboolean equip)
{
	DiversityBard *self;
	GSList *tmp_list;

	self = diversity_world_get_self(world);

	tmp_list = (GSList *) diversity_bard_get_equipments(self);
	while (tmp_list) {
		if (equip)
			diversity_bard_equip_equipment(self, tmp_list->data);
		else
			diversity_bard_unequip_equipment(self, tmp_list->data);

		tmp_list = tmp_list->next;
	}
}

static void diversity_log_handler(const gchar *log_domain, GLogLevelFlags log_level, const gchar *message, gpointer user_data)
{
	int priority;

	switch (log_level) {
	case G_LOG_LEVEL_ERROR:
		priority = LOG_CRIT;
		break;
	case G_LOG_LEVEL_CRITICAL:
		priority = LOG_ERR;
		break;
	case G_LOG_LEVEL_WARNING:
		priority = LOG_WARNING;
		break;
	case G_LOG_LEVEL_MESSAGE:
		priority = LOG_NOTICE;
		break;
	case G_LOG_LEVEL_DEBUG:
		priority = LOG_DEBUG;
		break;
	case G_LOG_LEVEL_INFO:
	default:
		priority = LOG_INFO;
		break;
	}

	syslog(priority, "%s", message);
}

static void logging_open(gboolean daemonize)
{
	if (daemonize)
		openlog(G_LOG_DOMAIN, LOG_CONS, LOG_DAEMON);
	else
		openlog(G_LOG_DOMAIN, LOG_CONS | LOG_PERROR, LOG_USER);

	g_log_set_handler(G_LOG_DOMAIN,
			G_LOG_LEVEL_MASK | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION,
			diversity_log_handler, NULL);
}

static void logging_close(void)
{
	closelog();
}

int main(int argc, char **argv)
{
	DiversityWorld *world = NULL;
	GError *error = NULL;
	gchar *confdir, *snap = NULL;
	
	confdir = (gchar *) g_get_home_dir();
	if (confdir) {
		confdir = g_build_filename(confdir, ".diversity", NULL);
		if (!g_file_test(confdir, G_FILE_TEST_IS_DIR)) {
			if (g_mkdir(confdir, 0777) < 0) {
				g_free(confdir);
				confdir = NULL;
			}
		}
	}
	
	if (confdir)
		snap = g_build_filename(confdir, "snapshot", NULL);

	if (!init(argc, argv, &error)) {
		g_printerr("failed to init: %s\n", error->message);

		goto fail;
	}
	
	if (!no_daemon) {
		g_printerr("FIXME: daemon mode is not implemented\n");
		no_daemon = TRUE;
	}

	logging_open(!no_daemon);

	g_type_init();

	world = g_object_new(DIVERSITY_TYPE_WORLD, NULL);

	if (snap)
		diversity_world_load_snapshot(world, snap);

	if (!backends_load(world, &error)) {
		g_warning("failed to load backends: %s",
				error->message);

		backends_unload(world);

		goto fail;
	}

	equip_equipments(world, TRUE);

	mainloop = g_main_loop_new(NULL, FALSE);
	g_main_loop_run(mainloop);
	g_main_loop_unref(mainloop);

	equip_equipments(world, FALSE);

	backends_unload(world);

	if (snap)
		diversity_world_save_snapshot(world, snap);

	g_object_unref(world);

	logging_close();

	g_free(snap);
	g_free(confdir);

	return 0;

fail:
	if (error)
		g_error_free(error);
	
	if (world)
		g_object_unref(world);

	g_free(snap);
	g_free(confdir);

	return 1;
}
