/* diversity-module.c - DiversityModule
 *
 * Copyright 2008 Openmoko, Inc.
 * Authored by Chia-I Wu <olv@openmoko.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <string.h> /* for strcmp */
#include <glib.h>
#include <gmodule.h>
#include <diversity-module.h>

#define DIVERSITY_MODULE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), DIVERSITY_TYPE_MODULE, DiversityModulePrivate))
typedef struct _DiversityModulePrivate DiversityModulePrivate;

struct _DiversityModulePrivate {
	gboolean builtin;

	gchar *path;
	GModule *module;

	DiversityModuleHandle *handle;
};

G_DEFINE_TYPE(DiversityModule, diversity_module, G_TYPE_OBJECT);

static void diversity_module_dispose(GObject *obj)
{
	((GObjectClass *) diversity_module_parent_class)->dispose(obj);
}

static void diversity_module_finalize(GObject *obj)
{
	DiversityModule *mod = DIVERSITY_MODULE(obj);
	DiversityModulePrivate *priv = DIVERSITY_MODULE_GET_PRIVATE(mod);

	if (!priv->builtin) {
		g_module_close(priv->module);
		g_free(priv->path);
	}

	((GObjectClass *) diversity_module_parent_class)->finalize(obj);
}

static void diversity_module_class_init(DiversityModuleClass *klass)
{
	GObjectClass *o_class = (GObjectClass *) klass;

	o_class->dispose = diversity_module_dispose;
	o_class->finalize = diversity_module_finalize;

	g_type_class_add_private(klass, sizeof(DiversityModulePrivate));

}

static void diversity_module_init(DiversityModule *mod)
{
	DiversityModulePrivate *priv = DIVERSITY_MODULE_GET_PRIVATE(mod);

	priv->path = NULL;
	priv->module = NULL;
}

DiversityModule *diversity_module_new(const char *path, GError **error)
{
	DiversityModule *mod;
	DiversityModulePrivate *priv;

	g_return_val_if_fail(path, NULL);

	mod = g_object_new(DIVERSITY_TYPE_MODULE, NULL);
	priv = DIVERSITY_MODULE_GET_PRIVATE(mod);

	priv->module = g_module_open(path, 0);
	if (!priv->module) {
		g_set_error(error, 0, 0, "failed to open %s: %s",
				path, g_module_error());

		g_object_unref(mod);

		return NULL;
	}

	if (!g_module_symbol(priv->module,
				DIVERSITY_MODULE_HANDLE_SYMBOL_NAME,
				(gpointer) &priv->handle) || !priv->handle) {
		g_set_error(error, 0, 0, "failed to get handle from %s: %s",
				path, g_module_error());

		g_module_close(priv->module);
		g_object_unref(mod);

		return NULL;
	}

	priv->path = g_strdup(path);

	return mod;
}

DiversityModule *diversity_module_new_from_handle(DiversityModuleHandle *handle)
{
	DiversityModule *mod;
	DiversityModulePrivate *priv;

	g_return_val_if_fail(handle, NULL);

	mod = g_object_new(DIVERSITY_TYPE_MODULE, NULL);
	priv = DIVERSITY_MODULE_GET_PRIVATE(mod);

	priv->builtin = TRUE;
	priv->handle = handle;

	return mod;
}

gboolean diversity_module_is_builtin(DiversityModule *mod)
{
	DiversityModulePrivate *priv = DIVERSITY_MODULE_GET_PRIVATE(mod);

	return priv->builtin;
}

DiversityModuleHandle *diversity_module_get_handle(DiversityModule *mod)
{
	DiversityModulePrivate *priv = DIVERSITY_MODULE_GET_PRIVATE(mod);

	return priv->handle;
}
